import numpy as np


def plot_line(ax, p1, p2, props="b-", linewidth=2):
    """
    Plot a line between points p1 and p2 on axis ax.

    """
    x1, y1, z1 = p1
    x2, y2, z2 = p2

    ax.plot([x1, x2], [y1, y2], [z1, z2], props, linewidth=linewidth)

    return ax


def plot_box(ax, p1, p2, props="b-", linewidth=2):
    """
    Plots a cube that spans between p1 and p2 on the given axis.

    Args:
      ax (matplolib axis): matplolib axis object
      p1 (tuple, list, np.ndarray): First cube point
        p1 is of length 3 (xmin, ymin, zmax).
      p2 (tuple, list, np.ndarray): Second cube point
        p2 is of length 3 (xmin, ymin, zmax).
      color (str): matplotlib color string
      linewidth (Real): matplotlib linewidth parameter

    """
    x1, y1, z1 = p1
    x2, y2, z2 = p2

    # Plot individual lines (edges) of a cube.
    plot_line(ax, (x1, y1, z1), (x2, y1, z1), props, linewidth=linewidth)
    plot_line(ax, (x1, y2, z1), (x2, y2, z1), props, linewidth=linewidth)
    plot_line(ax, (x1, y1, z2), (x2, y1, z2), props, linewidth=linewidth)
    plot_line(ax, (x1, y2, z2), (x2, y2, z2), props, linewidth=linewidth)

    plot_line(ax, (x1, y1, z1), (x1, y2, z1), props, linewidth=linewidth)
    plot_line(ax, (x2, y1, z1), (x2, y2, z1), props, linewidth=linewidth)
    plot_line(ax, (x1, y1, z2), (x1, y2, z2), props, linewidth=linewidth)
    plot_line(ax, (x2, y1, z2), (x2, y2, z2), props, linewidth=linewidth)

    plot_line(ax, (x1, y1, z1), (x1, y1, z2), props, linewidth=linewidth)
    plot_line(ax, (x2, y1, z1), (x2, y1, z2), props, linewidth=linewidth)
    plot_line(ax, (x1, y2, z1), (x1, y2, z2), props, linewidth=linewidth)
    plot_line(ax, (x2, y2, z1), (x2, y2, z2), props, linewidth=linewidth)

    return ax


def plane_line_intersection(n, p0, l, l0):
    """
    Computes an intersection point between line and a plane.

    """
    n, p0 = np.array(n), np.array(p0)
    l, l0 = np.array(l), np.array(l0)

    # Compensate for Special Case 1:
    # If both p0 and l0 are zero vectors, l0 is moved along the line vector,
    # so that it is a non-zero vector.
    if np.all(np.abs(p0) + np.abs(l0) == (0, 0, 0)):
        l0 += l

    if np.dot(l, n) == 0:
        # Line and plane are parallel to each other.
        if np.dot(p0-l0, n) == 0:
            # Line is contained in the plane.
            return False
        else:
            # No intersection
            return False
    else:
        d = np.dot(p0-l0, n) / np.dot(l, n)
        return tuple(d*l+l0)


def plane_plane_intersection(n1, p01, n2, p02):
    n1, p01 = np.array(n1), np.array(p01)
    n2, p02 = np.array(n2), np.array(p02)
    l = np.cross(n1, n2)

    if np.all(l == 0):
        # Planes are parallel to each other.
        return False
    else:
        # TODO
        pass


def box_line_intersection(pmin, pmax, l, l0):
    points = []
    for n in [(1, 0, 0), (0, 1, 0), (0, 0, 1)]:
        for p in [pmin, pmax]:
            points.append(plane_line_intersection(n, p, l, l0))

    # Remove False elements from the list.
    points = list(filter(lambda a: a is not False, points))

    # Find unique points inside box.
    points = points_inside_box(set(points), pmin, pmax)

    if len(points) != 2:
        return False
    else:
        return points


def points_inside_box(points, pmin, pmax):
    resulting_points = []
    for point in points:
        if pmin[0] <= point[0] <= pmax[0] and \
           pmin[1] <= point[1] <= pmax[1] and \
           pmin[2] <= point[2] <= pmax[2]:
            resulting_points.append(point)

    return tuple(resulting_points)
           



import os
import random
import pytest
import numpy as np
import discretisedfield as df
import matplotlib


class TestField(object):
    def setup(self):
        self.meshes = self.create_meshes()
        self.scalar_fs = self.create_scalar_fs()
        self.vector_fs = self.create_vector_fs()
        self.constant_values = [0, -5., np.pi,
                                1e-15, 1.2e12, random.random()]
        self.tuple_values = [(0, 0, 1),
                             (0, -5.1, np.pi),
                             [70, 1e15, 2*np.pi],
                             [5, random.random(), np.pi],
                             np.array([4, -1, 3.7]),
                             np.array([2.1, 0.0, -5*random.random()])]
        self.scalar_pyfuncs = self.create_scalar_pyfuncs()
        self.vector_pyfuncs = self.create_vector_pyfuncs()

    def create_meshes(self):
        p1_list = [(0, 0, 0),
                   (-5e-9, -8e-9, -10e-9),
                   (10, -5, -80)]
        p2_list = [(5e-9, 8e-9, 10e-9),
                   (11e-9, 4e-9, 4e-9),
                   (15, 10, 85)]
        cell_list = [(1e-9, 1e-9, 1e-9),
                     (1e-9, 2e-9, 1e-9),
                     (5, 5, 2.5)]

        meshes = []
        for i in range(len(p1_list)):
            mesh = df.Mesh(p1_list[i], p2_list[i], cell_list[i])
            meshes.append(mesh)
        return meshes

    def create_scalar_fs(self):
        scalar_fs = []
        for mesh in self.meshes:
            f = df.Field(mesh, dim=1)
            scalar_fs.append(f)
        return scalar_fs

    def create_vector_fs(self):
        vector_fs = []
        for mesh in self.meshes:
            f = df.Field(mesh, dim=3)
            vector_fs.append(f)
        return vector_fs

    def create_scalar_pyfuncs(self):
        f = [lambda c: 1,
             lambda c: -2.4,
             lambda c: -6.4e-15,
             lambda c: c[0] + c[1] + c[2] + 1,
             lambda c: (c[0]-1)**2 - c[1]+7 + c[2]*0.1,
             lambda c: np.sin(c[0]) + np.cos(c[1]) - np.sin(2*c[2])]

        return f

    def create_vector_pyfuncs(self):
        f = [lambda c: (1, 2, 0),
             lambda c: (-2.4, 1e-3, 9),
             lambda c: (c[0], c[1], c[2] + 100),
             lambda c: (c[0]+c[2]+10, c[1], c[2]+1),
             lambda c: (c[0]-1, c[1]+70, c[2]*0.1),
             lambda c: (np.sin(c[0]), np.cos(c[1]), -np.sin(2*c[2]))]

        return f

    def test_init(self):
        p1 = (0, -4, 11)
        p2 = (15, 10.1, 16.5)
        d = (1, 0.1, 0.5)
        mesh = df.Mesh(p1, p2, d)
        dim = 2
        name = "test_field"
        value = [1, 2]

        f = df.Field(mesh, dim=2, value=value, name=name)

        assert f.name == name

        assert np.all(f.f[:, :, :, 0] == value[0])
        assert np.all(f.f[:, :, :, 1] == value[1])

    def test_wrong_init(self):
        mesh = "wrong_mesh_string"
        with pytest.raises(TypeError):
            f = df.Field(mesh, dim=1, name="wrong_field")

    def test_set_with_constant(self):
        for value in self.constant_values:
            for f in self.scalar_fs + self.vector_fs:
                f.f = value

                # Check all values.
                assert np.all(f.f == value)

                # Check with sampling.
                assert np.all(f(f.mesh.random_point()) == value)

    def test_set_with_tuple_list_ndarray(self):
        for value in self.tuple_values:
            for f in self.vector_fs:
                f.normalisedto = 1
                f.f = value

                norm = (value[0]**2 + value[1]**2 + value[2]**2)**0.5
                for j in range(3):
                    c = f.mesh.random_point()
                    assert np.all(f.f[:, :, :, j] == value[j]/norm)
                    assert np.all(f(c)[j] == value[j]/norm)

    def test_set_from_callable(self):
        # Test scalar fs.
        for f in self.scalar_fs:
            for pyfun in self.scalar_pyfuncs:
                f.f = pyfun

                for j in range(10):
                    c = f.mesh.random_point()
                    expected_value = pyfun(f.mesh.cell_centre(c))
                    assert f(c) == expected_value

        # Test vector fields.
        for f in self.vector_fs:
            for pyfun in self.vector_pyfuncs:
                f.f = pyfun

                for j in range(10):
                    c = f.mesh.random_point()
                    expected_value = pyfun(f.mesh.cell_centre(c))
                    assert np.all(f(c) == expected_value)

    def test_set_exception(self):
        for f in self.vector_fs + self.scalar_fs:
            with pytest.raises(TypeError):
                f.f = "string"

    def test_average_scalar_field(self):
        value = -1e-3 + np.pi
        tol = 1e-12
        for f in self.scalar_fs:
            f.f = value
            average = f.average()

            assert abs(average[0] - value) < tol

    def test_average_vector_field(self):
        value = np.array([1.1, -2e-3, np.pi])
        tol = 1e-12
        for f in self.vector_fs:
            f.f = value
            average = f.average()

            assert abs(average[0] - value[0]) < tol
            assert abs(average[1] - value[1]) < tol
            assert abs(average[2] - value[2]) < tol

    def test_normalise(self):
        for f in self.vector_fs:
            for value in self.vector_pyfuncs + self.tuple_values:
                for norm_value in [1, 2.1, 50, 1e-3, np.pi]:
                    f.f = value
                    f.normalisedto = norm_value
                    f.normalise()

                    # Compute norm.
                    norm = 0
                    for j in range(f.dim):
                        norm += f.f[:, :, :, j]**2
                    norm = np.sqrt(norm)

                    assert norm.shape == (f.mesh.n[0],
                                          f.mesh.n[1],
                                          f.mesh.n[2])
                    diff = norm - norm_value

                    assert np.all(abs(norm - norm_value) < 1e-12)

    def test_normalise_scalar_field_exception(self):
        for f in self.scalar_fs:
            for value in self.scalar_pyfuncs:
                for norm_value in [1, 2.1, 50, 1e-3, np.pi]:
                    with pytest.raises(NotImplementedError):
                        f.normalisedto = norm_value
                        f.f = value

    def test_line_intersection(self):
        mesh = df.Mesh((0, 0, 0), (10e-9, 10e-9, 10e-9), (1e-9, 1e-9, 1e-9))
        field = df.Field(mesh, value=(1, 2, 3))

        d, v = field.line_intersection((0, 0, 1), (5e-9, 5e-9, 5e-9))

        assert len(d) == 100
        assert len(v) == 100

    def test_slice_field(self):
        for s in "xyz":
            for f in self.vector_fs + self.scalar_fs:
                if f.dim == 1:
                    funcs = self.scalar_pyfuncs
                elif f.dim == 3:
                    funcs = self.vector_pyfuncs

                for pyfun in funcs:
                    f.f = pyfun
                    point = f.mesh.centre()["xyz".find(s)]
                    data = f.slice_field(s, point)
                    a1, a2, f_slice, cs = data

                    if s == "x":
                        assert cs == (1, 2, 0)
                    elif s == "y":
                        assert cs == (0, 2, 1)
                    elif s == "z":
                        assert cs == (0, 1, 2)

                    tol = 1e-16
                    assert abs(a1[0] - (f.mesh.p1[cs[0]] +
                                        f.mesh.cell[cs[0]]/2.)) < tol
                    assert abs(a1[-1] - (f.mesh.p2[cs[0]] -
                                         f.mesh.cell[cs[0]]/2.)) < tol
                    assert abs(a2[0] - (f.mesh.p1[cs[1]] +
                                        f.mesh.cell[cs[1]]/2.)) < tol
                    assert abs(a2[-1] - (f.mesh.p2[cs[1]] -
                                         f.mesh.cell[cs[1]]/2.)) < tol
                    assert len(a1) == f.mesh.n[cs[0]]
                    assert len(a2) == f.mesh.n[cs[1]]
                    assert f_slice.shape == (f.mesh.n[cs[0]],
                                             f.mesh.n[cs[1]],
                                             f.dim)

                    for j in range(f.mesh.n[cs[0]]):
                        for k in range(f.mesh.n[cs[1]]):
                            c = list(f.mesh.centre())
                            c[cs[0]] = a1[j]
                            c[cs[1]] = a2[k]
                            c = tuple(c)

                            assert np.all(f_slice[j, k] == f(c))

    def test_slice_field_wrong_axis(self):
        for f in self.vector_fs + self.scalar_fs:
            if f.dim == 1:
                funcs = self.scalar_pyfuncs
            elif f.dim == 3:
                funcs = self.vector_pyfuncs

            for pyfun in funcs:
                f.f = pyfun
                point = f.mesh.centre()[0]
                with pytest.raises(ValueError):
                    data = f.slice_field("xy", point)
                    data = f.slice_field("xyz", point)
                    data = f.slice_field("zy", point)
                    data = f.slice_field("string", point)
                    data = f.slice_field("point", point)

    def test_slice_field_wrong_point(self):
        for s in "xyz":
            for f in self.vector_fs + self.scalar_fs:
                if f.dim == 1:
                    funcs = self.scalar_pyfuncs
                elif f.dim == 3:
                    funcs = self.vector_pyfuncs

                for pyfun in funcs:
                    f.f = pyfun
                    point = f.mesh.centre()["xyz".find(s)] + 100
                    with pytest.raises(ValueError):
                        data = f.slice_field(s, point)

                    point = f.mesh.centre()["xyz".find(s)] - 100
                    with pytest.raises(ValueError):
                        data = f.slice_field(s, point)

    def test_plot_slice_vector_field(self):
        figname = "test_slice_plot_figure.pdf"
        value = (1e-3 + np.pi, -5, 6)
        for f in self.vector_fs:
            f.normalisedto = 1
            f.f = value
            point = f.mesh.centre()["xyz".find("y")]
            fig = f.plot_slice("y", point, axes=True)
            fig = f.plot_slice("y", point, axes=False)

    def test_plot_slice_vector_field_exception(self):
        value = (0, 0, 1)
        for f in self.vector_fs:
            f.normalisedto = 1
            f.f = value
            point = f.mesh.centre()["xyz".find("z")]
            with pytest.raises(ValueError):
                fig = f.plot_slice("z", point)

    def test_write_read_oommf_file_text(self):
        tol = 1e-12
        filename = "test_write_oommf_file_text.omf"
        value = (1e-3 + np.pi, -5, 6)
        for f in self.vector_fs:
            f.f = value
            f.write_oommf_file(filename)

            f_loaded = df.read_oommf_file(filename)

            assert f.mesh.p1 == f_loaded.mesh.p1
            assert f.mesh.p2 == f_loaded.mesh.p2
            assert f.mesh.cell == f_loaded.mesh.cell
            assert f.mesh.cell == f_loaded.mesh.cell
            assert np.all(abs(f.f - f_loaded.f) < tol)

            os.system("rm {}".format(filename))

    def test_write_read_oommf_file(self):
        tol = 1e-12
        filename = "test_write_oommf_file_binary.omf"
        value = (1e-3 + np.pi, -5, 6)
        for f in self.vector_fs:
            f.f = value
            f.write_oommf_file(filename, 'binary')

            f_loaded = df.read_oommf_file(filename)

            assert f.mesh.p1 == f_loaded.mesh.p1
            assert f.mesh.p2 == f_loaded.mesh.p2
            assert f.mesh.cell == f_loaded.mesh.cell
            assert f.mesh.cell == f_loaded.mesh.cell
            assert np.all(abs(f.f - f_loaded.f) < tol)

            os.system("rm {}".format(filename))



import numpy as np
import joommfutil.typesystem as ts
import discretisedfield as df
import discretisedfield.util as dfu
import matplotlib.pyplot as plt
import struct

# TODO: rename f
# TODO: sample arbitrary plane


@ts.typesystem(mesh=ts.TypedAttribute(expected_type=df.Mesh),
               dim=ts.UnsignedInt,
               name=ts.ObjectName)
class Field(object):

    def __init__(self, mesh, dim=3, value=0,
                 normalisedto=None, name="field"):
        """Class for analysing and manipulating Finite Difference (FD) fields.

        This class provides the functionality for:
          - creating FD scalar and vector fields,
          - plotting FD fields,
          - computing common values characteristic to FD fields, and
          - saving FD fields in common file formats.

        Args:
          mesh (Mesh): Finite difference mesh.
          dim (Optional[int]): The value dimensionality. Defaults to 3.
          value (Optional): Finite difference field value. Defaults to 0.
            For the possible types of value argument, refer to the f.setter method.
            If no value argument is provided, a zero field is initialised.
          normalisedto (Optional[Real]): vector field norm
          name (Optional[str]): Field name

        Attributes:
          mesh (Mesh): Finite difference mesh

          dim (int): The value dimensionality.

          normalisedto (Real): Vector field norm

          name (str): Field name

          f (np.ndarray): Field value - a four-dimensional numpy array

        """
        self.mesh = mesh
        self.dim = dim
        self.normalisedto = normalisedto
        self.name = name
        self.f = value  # calls setter method

    @property
    def f(self):
        return self._f

    @f.setter
    def f(self, value):
        """Set the field value.

        This method sets the field values at all finite difference
        domain cells.

        Args:
          value: This argument can be an integer, float, tuple, list,
            np.ndarray, or Python function.

        """
        if not hasattr(self, "f"):
            self._f = np.zeros((self.mesh.n[0],
                                self.mesh.n[1],
                                self.mesh.n[2],
                                self.dim))

        if isinstance(value, (int, float)):
            self._f.fill(value)
        elif isinstance(value, (tuple, list, np.ndarray)):
            for i in range(self.dim):
                self._f[:, :, :, i].fill(value[i])
        elif callable(value):
            for i, p in self.mesh.cells():
                self._f[i[0], i[1], i[2], :] = value(p)
        else:
            raise TypeError("Cannot set field using {}.".format(type(value)))

        if self.normalisedto is not None:
            self.normalise()

    def normalise(self):
        """Normalise the vector field to self.normalisedto value."""
        if self.dim == 1:
            raise NotImplementedError("Normalisation is supported only "
                                      "for vector fields.")
        norm = np.linalg.norm(self.f, axis=3)
        for i in range(self.dim):
            self.f[:, :, :, i] = self.normalisedto*self.f[:, :, :, i]/norm

    def __call__(self, p):
        """Sample the field at point p.

        Args:
          p (tuple): point coordinate at which the field is sampled

        Returns:
          Field value in cell containing point p

        """
        i, j, k = self.mesh.point2index(p)
        return self.f[i, j, k]

    def average(self):
        """Compute the finite difference field average.

        Returns:
          Finite difference field average.

        """
        return tuple(self.f.mean(axis=(0, 1, 2)))

    def line_intersection(self, l, l0, n=100):
        """
        Slice field along the line defined with l and l0.
        """
        ds, points, values = [], [], []
        for d, p in self.mesh.line_intersection(l, l0, n=n):
            ds.append(d)
            points.append(p)
            values.append(self.__call__(p))

        return ds, values

    def plot_line_intersection(self, l, l0, n=100):
        # Plot schematic representation of intersection.
        fig = plt.figure()
        ax = fig.add_subplot(211, projection="3d")
        ax.set_aspect("equal")

        dfu.plot_box(ax, self.mesh.pmin, self.mesh.pmax)
        p1, p2 = dfu.box_line_intersection(
            self.mesh.pmin, self.mesh.pmax, l, l0)
        dfu.plot_line(ax, p1, p2, props="ro-")
        ax.set(xlabel=r"$x$", ylabel=r"$y$", zlabel=r"$z$")

        # Plot field along line
        ax = fig.add_subplot(212)
        d, v = self.line_intersection(l, l0, n=n)
        ax.set(xlabel=r"$d$", ylabel=r"$v$")
        ax.grid()
        ax.plot(d, v)

        return fig

    def slice_field(self, axis, point):
        """Returns the field slice.

        This method returns the values of a finite difference field
        on the plane perpendicular to the 'axis' at 'point'.

        Args:
          axis (str): An axis to which the sampling plane is perpendicular to.
          point (int/float): The coordinate on axis at which the field is
            sampled.

        Returns:
          A 4 element tuple containing:

            - Axis 1 coodinates
            - Axis 2 coodinates
            - np.ndarray of field values on the plane
            - coordinate system details

        """
        if axis == "x":
            slice_num = 0
            axes = (1, 2)
        elif axis == "y":
            slice_num = 1
            axes = (0, 2)
        elif axis == "z":
            slice_num = 2
            axes = (0, 1)
        else:
            raise ValueError("Axis not properly defined.")

        if self.mesh.p1[slice_num] <= point <= self.mesh.p2[slice_num]:
            axis1_indices = np.arange(0, self.mesh.n[axes[0]])
            axis2_indices = np.arange(0, self.mesh.n[axes[1]])

            axis1_coords = np.zeros(len(axis1_indices))
            axis2_coords = np.zeros(len(axis2_indices))

            sample_centre = list(self.mesh.centre())
            sample_centre[slice_num] = point
            sample_centre = tuple(sample_centre)

            slice_index = self.mesh.point2index(sample_centre)[slice_num]

            field_slice = np.zeros([self.mesh.n[axes[0]],
                                    self.mesh.n[axes[1]],
                                    self.dim])
            for j in axis1_indices:
                for k in axis2_indices:
                    i = [0, 0, 0]
                    i[slice_num] = slice_index
                    i[axes[0]] = j
                    i[axes[1]] = k
                    i = tuple(i)

                    coord = self.mesh.index2point(i)

                    axis1_coords[j] = coord[axes[0]]
                    axis2_coords[k] = coord[axes[1]]

                    field_slice[j, k, :] = self.f[i[0], i[1], i[2], :]
            coord_system = (axes[0], axes[1], slice_num)

        else:
            raise ValueError("Point {} outside the domain.".format(point))

        return axis1_coords, axis2_coords, field_slice, coord_system

    def plot_slice(self, axis, point, xsize=10, axes=True, grid=True):
        """Plot the field slice.

        This method plots the field slice that is obtained
        using slice_field method.

        Args:
          axis (str): An axis to which the sampling plane is perpendicular to.
          point (int/float): The coordinate axis at which the field
            is sampled.
          xsize (Optional[int/float]): The horizontal size of a plot.
          grid (Optional[bool]): If True, grid is shown in the plot.

        Returns:
          matplotlib figure.

        """
        a1, a2, field_slice, coord_system = self.slice_field(axis, point)

        # Vector field.
        if self.dim == 3:
            pm = self._prepare_for_quiver(a1, a2, field_slice, coord_system)

            if np.allclose(pm[:, 2], 0) and np.allclose(pm[:, 3], 0):
                raise ValueError("Vector plane components are zero.")
            else:
                ysize = xsize*(self.mesh.l[coord_system[1]] /
                               self.mesh.l[coord_system[0]])
                fig = plt.figure(figsize=(xsize, ysize))
                plt.quiver(pm[:, 0], pm[:, 1], pm[:, 2], pm[:, 3], pm[:, 4])
                plt.xlim([self.mesh.p1[coord_system[0]],
                          self.mesh.p2[coord_system[0]]])
                plt.ylim([self.mesh.p1[coord_system[1]],
                          self.mesh.p2[coord_system[1]]])
                if axes:
                    plt.xlabel("xyz"[coord_system[0]] + " (m)")
                    plt.ylabel("xyz"[coord_system[1]] + " (m)")
                    plt.title("xyz"[coord_system[2]] + " slice")
                if not axes:
                    plt.axis("off")
                if grid:
                    plt.grid()

        return fig

    def _prepare_for_quiver(self, a1, a2, field_slice, coord_system):
        """Generate arrays for plotting quiver plot."""
        nel = self.mesh.n[coord_system[0]]*self.mesh.n[coord_system[1]]
        plot_matrix = np.zeros([nel, 5])

        counter = 0
        for j in range(self.mesh.n[coord_system[0]]):
            for k in range(self.mesh.n[coord_system[1]]):
                entry = [a1[j], a2[k],
                         field_slice[j, k, coord_system[0]],
                         field_slice[j, k, coord_system[1]],
                         field_slice[j, k, coord_system[2]]]
                plot_matrix[counter, :] = np.array(entry)
                counter += 1

        return plot_matrix

    def write_oommf_file(self, filename, datatype='text'):
        """Write the FD field to the OOMMF (omf, ohf) file.
        This method writes all necessary data to the omf or ohf file,
        so that it can be read by OOMMF.
        Args:
          filename (str): filename including extension
          type(str): Either 'text' or 'binary'
        Example:
        .. code-block:: python
          >>> from oommffield import Field
          >>> field = Field((0, 0, 0), (5, 4, 3), (1, 1, 1))
          >>> field.set((1, 0, 5))
          >>> field.write_oommf_file('fdfield.omf')
        """
        oommf_file = open(filename, 'w')

        # Define header lines.
        header_lines = ['OOMMF OVF 2.0',
                        '',
                        'Segment count: 1',
                        '',
                        'Begin: Segment',
                        'Begin: Header',
                        '',
                        'Title: Field generated omf file',
                        'Desc: File generated by Field class',
                        'meshunit: m',
                        'meshtype: rectangular',
                        'xbase: {}'.format(self.mesh.cell[0]),
                        'ybase: {}'.format(self.mesh.cell[1]),
                        'zbase: {}'.format(self.mesh.cell[2]),
                        'xnodes: {}'.format(self.mesh.n[0]),
                        'ynodes: {}'.format(self.mesh.n[1]),
                        'znodes: {}'.format(self.mesh.n[2]),
                        'xstepsize: {}'.format(self.mesh.cell[0]),
                        'ystepsize: {}'.format(self.mesh.cell[1]),
                        'zstepsize: {}'.format(self.mesh.cell[2]),
                        'xmin: {}'.format(self.mesh.p1[0]),
                        'ymin: {}'.format(self.mesh.p1[1]),
                        'zmin: {}'.format(self.mesh.p1[2]),
                        'xmax: {}'.format(self.mesh.p2[0]),
                        'ymax: {}'.format(self.mesh.p2[1]),
                        'zmax: {}'.format(self.mesh.p2[2]),
                        'valuedim: {}'.format(self.dim),
                        'valuelabels: Magnetization_x Magnetization_y Magnetization_z',
                        'valueunits: A/m A/m A/m',
                        '',
                        'End: Header',
                        '']

        if datatype == 'binary':
            header_lines.append('Begin: Data Binary 8')
            footer_lines = ['End: Data Binary 8',
                            'End: Segment']
        if datatype == 'text':
            header_lines.append('Begin: Data Text')
            footer_lines = ['End: Data Text',
                            'End: Segment']

        # Write header lines to OOMMF file.
        for line in header_lines:
            if line == '':
                oommf_file.write('#\n')
            else:
                oommf_file.write('# ' + line + '\n')
        if datatype == 'binary':
            # Close the file and reopen with binary write
            # appending to end of file.
            oommf_file.close()
            oommf_file = open(filename, 'ab')
            # Add the 8 bit binary check value that OOMMF uses
            packarray = [123456789012345.0]
            # Write data lines to OOMMF file.
            for iz in range(self.mesh.n[2]):
                for iy in range(self.mesh.n[1]):
                    for ix in range(self.mesh.n[0]):
                        [packarray.append(vi) for vi in self.f[ix, iy, iz, :]]

            v_binary = struct.pack('d'*len(packarray), *packarray)
            oommf_file.write(v_binary)
            oommf_file.close()
            oommf_file = open(filename, 'a')

        else:
            for iz in range(self.mesh.n[2]):
                for iy in range(self.mesh.n[1]):
                    for ix in range(self.mesh.n[0]):
                        v = [str(vi) for vi in self.f[ix, iy, iz, :]]
                        for vi in v:
                            oommf_file.write(' ' + vi)
                        oommf_file.write('\n')
        # Write footer lines to OOMMF file.
        for line in footer_lines:
            oommf_file.write('# ' + line + '\n')

        # Close the file.
        oommf_file.close()


def read_oommf_file(filename, normalisedto=None, name='unnamed'):
    try:
        f = open(filename)

        if 'Begin: Data Text' in f.read():
            field = read_oommf_file_text(filename, name)
        else:
            field = read_oommf_file_binary(filename, name)
    except UnicodeDecodeError:
        field = read_oommf_file_binary(filename, name)

    field.normalisedto = normalisedto
    if normalisedto is not None:
        field.normalise()

    return field


def read_oommf_file_text(filename, name="unnamed"):
    """Read the OOMMF file and create an Field object.
    Args:
      filename (str): OOMMF file name
      name (str): name of the Field object
    Return:
      Field object.
    Example:
        .. code-block:: python
          from oommffield import read_oommf_file
          oommf_filename = 'vector_field.omf'
          field = read_oommf_file(oommf_filename, name='magnetisation')
    """
    # Open and read the file.
    f = open(filename, 'r')
    lines = f.readlines()
    f.close()

    # Load metadata.
    dic = {'xmin': None, 'ymin': None, 'zmin': None,
           'xmax': None, 'ymax': None, 'zmax': None,
           'xstepsize': None, 'ystepsize': None, 'zstepsize': None,
           'xbase': None, 'ybase': None, 'zbase': None,
           'xnodes': None, 'ynodes': None, 'znodes': None,
           'valuedim': None}

    for line in lines[0:50]:
        for key in dic.keys():
            if line.find(key) != -1:
                dic[key] = float(line.split()[2])

    p1 = (dic["xmin"], dic["ymin"], dic["zmin"])
    p2 = (dic["xmax"], dic["ymax"], dic["zmax"])
    d = (dic["xstepsize"], dic["ystepsize"], dic["zstepsize"])
    cbase = (dic["xbase"], dic["ybase"], dic["zbase"])
    n = (int(round(dic["xnodes"])),
         int(round(dic["ynodes"])),
         int(round(dic["znodes"])))
    dim = int(dic["valuedim"])
    mesh = df.Mesh(p1, p2, d, name=name)
    field = Field(mesh, dim, value=(1, 1, 1), name=name)

    for j in range(len(lines)):
        if lines[j].find('Begin: Data Text') != -1:
            data_first_line = j+1

    counter = 0
    for i, _ in mesh.cells():
        line_data = lines[data_first_line+counter]
        value = [float(vi) for vi in line_data.split()]
        field.f[i[0], i[1], i[2], :] = value

        counter += 1

    return field


def read_oommf_file_binary(filename, name='unnamed'):
    """Read the OOMMF file and create an Field object.
    Args:
      filename (str): OOMMF file name
      name (str): name of the Field object
    Return:
      Field object.
    Example:
        .. code-block:: python
          from oommffield import read_oommf_file
          oommf_filename = 'vector_field.omf'
          field = read_oommf_file(oommf_filename, name='magnetisation')
    """
    # Open and read the file.
    with open(filename, 'rb') as f:
        file = f.read()
        lines = file.split(b'\n')

    # Load metadata.
    dic = {'xmin': None, 'ymin': None, 'zmin': None,
           'xmax': None, 'ymax': None, 'zmax': None,
           'xstepsize': None, 'ystepsize': None, 'zstepsize': None,
           'xbase': None, 'ybase': None, 'zbase': None,
           'xnodes': None, 'ynodes': None, 'znodes': None,
           'valuedim': None}

    for line in lines[0:50]:
        for key in dic.keys():
            if line.find(bytes(key, 'utf-8')) != -1:
                dic[key] = float(line.split()[2])

    p1 = (dic["xmin"], dic["ymin"], dic["zmin"])
    p2 = (dic["xmax"], dic["ymax"], dic["zmax"])
    d = (dic["xstepsize"], dic["ystepsize"], dic["zstepsize"])
    cbase = (dic["xbase"], dic["ybase"], dic["zbase"])
    n = (int(round(dic["xnodes"])),
         int(round(dic["ynodes"])),
         int(round(dic["znodes"])))
    dim = int(dic["valuedim"])

    mesh = df.Mesh(p1, p2, d, name=name)
    field = Field(mesh, dim, value=(1, 1, 1), name=name)

    binary_header = b'# Begin: Data Binary '
    # Here we find the start and end points of the
    # binary data, in terms of byte position.
    data_start = file.find(binary_header)
    header = file[data_start:data_start + len(binary_header) + 1]
    if b'8' in header:
        bytesize = 8
    elif b'4' in header:
        bytesize = 4

    data_start += len(b'# Begin: Data Binary 8\n')
    data_end = file.find(b'# End: Data Binary ')
    if bytesize == 4:
        listdata = list(struct.iter_unpack('@f', file[data_start:data_end]))
        try:
            assert listdata[0] == 1234567.0
        except:
            raise AssertionError('Something has gone wrong'
                                 ' with reading Binary Data')
    elif bytesize == 8:
        listdata = list(struct.iter_unpack('@d', file[data_start:data_end]))
        try:
            assert listdata[0][0] == 123456789012345.0
        except:
            raise AssertionError('Something has gone wrong'
                                 ' with reading Binary Data')

    counter = 1
    for i, _ in mesh.cells():
        value = (listdata[counter][0],
                 listdata[counter+1][0],
                 listdata[counter+2][0])
        field.f[i[0], i[1], i[2], :] = value

        counter += 3

    return field



import numpy as np
import discretisedfield.util as dfu


def test_plane_line_intersection():
    assert dfu.plane_line_intersection((0, 1, 0), (0, 1, 0),
                                       (0, 1, 0), (0, 0, 0)) == (0, 1, 0)
    assert dfu.plane_line_intersection((1, 0, 0), (1, 0, 0),
                                       (1, 1, 1), (0, 0, 0)) == (1, 1, 1)
    assert dfu.plane_line_intersection((1, 0, 0), (1, 0, 0),
                                       (1, -3, 1), (0, 0, 0)) == (1, -3, 1)

    # Special case 1: The intersection point should be (0, 0, 0),
    # however the equation results in d=0, implying there is not
    # a single intersection point.
    assert dfu.plane_line_intersection((1, 0, 0), (0, 0, 0),
                                       (1, 0, 0), (0, 0, 0)) == (0, 0, 0)
    assert dfu.plane_line_intersection((0, 1, 0), (0, 0, 0),
                                       (0, 1, 0), (0, 0, 0)) == (0, 0, 0)
    assert dfu.plane_line_intersection((0, 0, 1), (0, 0, 0),
                                       (0, 0, 1), (0, 0, 0)) == (0, 0, 0)
    assert dfu.plane_line_intersection((1, 0, 0), (0, 0, 0),
                                       (1, 1, 1), (0, 0, 0)) == (0, 0, 0)

    # Special case 2: n and l are perpendicular to each other,
    # which means that line and plane are parallel to each other.
    assert dfu.plane_line_intersection((1, 0, 0), (0, 0, 0),
                                       (0, 0, 1), (0, 0, 0)) is False
    assert dfu.plane_line_intersection((0, 1, 0), (0, 0, 0),
                                       (0, 0, 1), (0, 0, 0)) is False
    assert dfu.plane_line_intersection((1, 0, 0), (5, 1, 3),
                                       (0, 0, 1), (11, 9, -6)) is False


def test_box_line_intersection():
    points = dfu.box_line_intersection((0, 0, 0), (1, 1, 1),
                                       (1, 0, 0), (0, 0, 0))
    assert (0, 0, 0) in points
    assert (1, 0, 0) in points

    points = dfu.box_line_intersection((0, 0, 0), (1, 1, 1),
                                       (0, 1, 0), (0, 0, 0))
    assert (0, 0, 0) in points
    assert (0, 1, 0) in points

    points = dfu.box_line_intersection((0, 0, 0), (1, 1, 1),
                                       (0, 0, 1), (0, 0, 0))
    assert (0, 0, 0) in points
    assert (0, 0, 1) in points

    points = dfu.box_line_intersection((0, 0, 0), (5, 1, 1),
                                       (1, 0, 0), (0, 0, 0))
    assert (0, 0, 0) in points
    assert (5, 0, 0) in points

    points = dfu.box_line_intersection((0, 0, 0), (1, 1, 1),
                                       (1, 1, 1), (0, 0, 0))
    assert (0, 0, 0) in points
    assert (1, 1, 1) in points

    points = dfu.box_line_intersection((0, 0, 0), (1, 1, 1),
                                       (2, 2, 2), (5, 5, 5))
    assert (0, 0, 0) in points
    assert (1, 1, 1) in points

    points = dfu.box_line_intersection((0, 0, 0), (1, 1, 1),
                                       (0, 0, 1), (0.99, 0.5, 0.5))
    assert (0.99, 0.5, 0) in points
    assert (0.99, 0.5, 1) in points

    # Special case 1: Line intersection outside the box
    points = dfu.box_line_intersection((0, 0, 0), (1, 1, 1),
                                       (1, 0, 0), (5, 5, 5))
    assert points is False

    points = dfu.box_line_intersection((0, 0, 0), (1, 1, 1),
                                       (0, 0, 1), (1.01, 0.5, 0.5))
    assert points is False

    # Special case 2: Only one intersection point
    points = dfu.box_line_intersection((0, 0, 0), (-1, -1, 1),
                                       (1, 1, 1), (0, 0, 0))
    assert points is False



# use most conservative backend 
import matplotlib
matplotlib.use("agg")

from .mesh import Mesh
from .field import Field, read_oommf_file


def test():
    import pytest  # pragma: no cover
    pytest.main(["-v", "--pyargs", "discretisedfield"])  # pragma: no cover



import pytest
import matplotlib
import numpy as np
import discretisedfield as df


class TestMesh(object):
    def setup(self):
        self.valid_args = [[(0, 0, 0),
                            (5, 5, 5),
                            [1, 1, 1]],
                           [(0, 0, 0),
                            (5e-9, 5e-9, 5e-9),
                            (1e-9, 1e-9, 1e-9)],
                           [(-1.5e-9, -5e-9, 0),
                            (1.5e-9, -15e-9, -160e-9),
                            (0.3e-9, 1e-9, 1e-9)],
                           [(-1.5e-9, -5e-9, -5e-9),
                            np.array((0, 0, 0)),
                            (0.5e-9, 1e-9, 5e-9)],
                           [[0, 5e-6, 0],
                            (-1.5e-9, -5e-9, -5e-9),
                            (0.5e-9, 1e-9, 5e-9)],
                           [(0, 0, 0),
                            (500e-9, 125e-9, 3e-9),
                            (2.5e-9, 2.5e-9, 3e-9)]]

        self.invalid_args = [[(0, 0, 0),
                              (5, 5, 5),
                              (-1, 1, 1)],
                             ["1",
                              (1, 1, 1),
                              (0, 0, 1e-9)],
                             [(-1.5e-9, -5e-9, "a"),
                              (1.5e-9, 15e-9, 16e-9),
                              (5, 1, -1e-9)],
                             [(-1.5e-9, -5e-9, 0),
                              (1.5e-9, 16e-9),
                              (0.1e-9, 0.1e-9, 1e-9)],
                             [(-1.5e-9, -5e-9, 0),
                              (1.5e-9, 15e-9, 1+2j),
                              (5, 1, 1e-9)],
                             ["string",
                              (5, 1, 1e-9),
                              "string"],
                             [(-1.5e-9, -5e-9, 0),
                              (1.5e-9, 15e-9, 16e-9),
                              1]]

    def test_simple_init(self):
        p1 = (0, -4, 16.5)
        p2 = (15, 10.1, 11)
        cell = (1, 0.1, 0.5)
        name = "test_mesh"
        mesh = df.Mesh(p1, p2, cell, name=name)

        assert isinstance(mesh.p1, tuple)
        assert mesh.p1 == p1
        assert isinstance(mesh.p2, tuple)
        assert mesh.p2 == p2
        assert isinstance(mesh.cell, tuple)
        assert mesh.cell == cell
        assert isinstance(mesh.name, str)
        assert mesh.name == name
        assert isinstance(mesh.pmin, tuple)
        assert mesh.pmin == (0, -4, 11)
        assert isinstance(mesh.pmax, tuple)
        assert mesh.pmax == (15, 10.1, 16.5)

        assert isinstance(mesh.l, tuple)
        assert mesh.l[0] == 15 - 0
        assert mesh.l[1] == 10.1 - (-4)
        assert mesh.l[2] == 16.5 - 11

        assert isinstance(mesh.n, tuple)
        assert mesh.n[0] == (15 - 0) / 1.0
        assert mesh.n[1] == (10.1 - (-4)) / 0.1
        assert mesh.n[2] == (16.5 - 11) / 0.5
        for i in range(3):
            assert isinstance(mesh.n[i], int)

    def test_init_valid_args(self):
        for arg in self.valid_args:
            p1, p2, cell = arg
            mesh = df.Mesh(p1, p2, cell)

            assert isinstance(mesh.p1, tuple)
            assert mesh.p1 == tuple(p1)
            assert isinstance(mesh.p2, tuple)
            assert mesh.p2 == tuple(p2)
            assert isinstance(mesh.cell, tuple)
            assert mesh.cell == tuple(cell)
            assert isinstance(mesh.name, str)
            assert mesh.name == "mesh"  # default name value
            assert isinstance(mesh.l, tuple)
            assert isinstance(mesh.n, tuple)
            for i in range(3):
                assert isinstance(mesh.n[i], int)

    def test_init_invalid_args(self):
        for arg in self.invalid_args:
            p1, p2, cell = arg
            with pytest.raises(TypeError):
                mesh = df.Mesh(p1, p2, cell)

    def test_zero_domain_edge(self):
        # Exception is raised by the descriptor
        p1 = (0, 100e-9, 1e-9)
        p2 = (150e-9, 100e-9, 6e-9)
        cell = (1e-9, 1e-9, 1e-9)
        with pytest.raises(TypeError):
            mesh = df.Mesh(p1, p2, cell)

        p1 = (0, 100e-9, 0)
        p2 = (150e-9, 101e-9, 0)
        cell = (1e-9, 1e-9, 1e-9)
        with pytest.raises(TypeError):
            mesh = df.Mesh(p1, p2, cell)

    def test_init_d_not_multiple_of_l(self):
        p1 = (0, -4, 11)
        p2 = (15, 20, 16e-9)
        cell = (1, 5, 1e-9)
        with pytest.raises(ValueError):
            mesh = df.Mesh(p1, p2, cell)

        p1 = (0, 0, 0)
        p2 = (500e-9, 125e-9, 3e-9)
        cell = (2.5e-9, 2.5e-9, 2.5e-9)
        with pytest.raises(ValueError):
            mesh = df.Mesh(p1, p2, cell)

    def test_init_invalid_name(self):
        # Exception raised by the descriptor
        p1 = (0, 0, 0)
        p2 = (500e-9, 125e-9, 3e-9)
        cell = (2.5e-9, 2.5e-9, 3e-9)
        for name in ["mesh name", "2name", " ", 5]:
            with pytest.raises(TypeError):
                mesh = df.Mesh(p1, p2, cell, name=name)

    def test_repr(self):
        p1 = (-1, -4, 11)
        p2 = (15, 10.1, 12.5)
        cell = (1, 0.1, 0.5)
        name = "meshname"
        mesh = df.Mesh(p1, p2, cell, name=name)

        rstr = ("Mesh(p1=(-1, -4, 11), p2=(15, 10.1, 12.5), "
                "cell=(1, 0.1, 0.5), name=\"meshname\")")

        assert repr(mesh) == rstr

    def test_index2point(self):
        p1 = (15, -4, 12.5)
        p2 = (-1, 10.1, 11)
        cell = (1, 0.1, 0.5)
        mesh = df.Mesh(p1, p2, cell)

        assert mesh.index2point((0, 0, 0)) == (-0.5, -3.95, 11.25)
        assert mesh.index2point((5, 10, 1)) == (4.5, -2.95, 11.75)
        assert mesh.index2point((15, 140, 2)) == (14.5, 10.05, 12.25)

    def test_index2point_exception(self):
        p1 = (-1, -4, 11)
        p2 = (15, 10.1, 12.5)
        cell = (1, 0.1, 0.5)
        mesh = df.Mesh(p1, p2, cell)

        # Correct minimum index
        assert isinstance(mesh.index2point((0, 0, 0)), tuple)
        # Below minimum index
        with pytest.raises(ValueError):
            mesh.index2point((-1, 0, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, -1, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, 0, -1))

        # Correct maximum index
        assert isinstance(mesh.index2point((15, 140, 2)), tuple)
        # Above maximum index
        with pytest.raises(ValueError):
            mesh.index2point((16, 0, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, 141, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, 0, 3))

    def test_point2index(self):
        p1 = (-10e-9, -5e-9, 10e-9)
        p2 = (10e-9, 5e-9, 0)
        cell = (1e-9, 5e-9, 1e-9)
        mesh = df.Mesh(p1, p2, cell)

        # (0, 0, 0) cell
        assert mesh.point2index((-10e-9, -5e-9, 0)) == (0, 0, 0)
        assert mesh.point2index((-9.5e-9, -2.5e-9, 0.5e-9)) == (0, 0, 0)
        assert mesh.point2index((-9.01e-9, -0.1e-9, 0.9e-9)) == (0, 0, 0)

        # (19, 1, 9) cell
        assert mesh.point2index((10e-9, 5e-9, 10e-9)) == (19, 1, 9)
        assert mesh.point2index((9.5e-9, 2.5e-9, 9.5e-9)) == (19, 1, 9)
        assert mesh.point2index((9.1e-9, 0.1e-9, 9.1e-9)) == (19, 1, 9)

        # vicinity of (0, 0, 0) point
        assert mesh.point2index((1e-16, 1e-16, 1e-16)) == (10, 1, 0)
        assert mesh.point2index((-1e-16, -1e-16, 1e-16)) == (9, 0, 0)

    def test_point2index_exception(self):
        p1 = (-10, 5, 0)
        p2 = (10, -5, 10e-9)
        cell = (1, 5, 1e-9)
        mesh = df.Mesh(p1, p2, cell)

        tol = 1e-12
        with pytest.raises(ValueError):
            mesh.point2index((-10-tol, 0, 5))
        with pytest.raises(ValueError):
            mesh.point2index((-5, -5-tol, 5))
        with pytest.raises(ValueError):
            mesh.point2index((-5, 0, -tol))
        with pytest.raises(ValueError):
            mesh.point2index((10+tol, 0, 5))
        with pytest.raises(ValueError):
            mesh.point2index((6, 5+tol, 5))
        with pytest.raises(ValueError):
            mesh.point2index((0, 0, 10e-9+tol))

    def test_index2point_point2index_mutually_inverse(self):
        p1 = (15, -4, 12.5)
        p2 = (-1, 10.1, 11)
        cell = (1, 0.1, 0.5)
        mesh = df.Mesh(p1, p2, cell)

        for i in [(-0.5, -3.95, 11.25), (14.5, 10.05, 12.25)]:
            assert isinstance(mesh.index2point(mesh.point2index(i)), tuple)
            assert mesh.index2point(mesh.point2index(i)) == i

        for i in [(1, 0, 0), (0, 1, 0), (0, 0, 1)]:
            assert isinstance(mesh.point2index(mesh.index2point(i)), tuple)
            assert mesh.point2index(mesh.index2point(i)) == i

    def test_cell_centre(self):
        p1 = (500e-9, 125e-9, 3e-9)
        p2 = (0, 0, 0)
        cell = (10e-9, 5e-9, 1e-9)
        mesh = df.Mesh(p1, p2, cell)

        assert isinstance(mesh.cell_centre((500e-9, 0, 0)), tuple)
        assert mesh.cell_centre((0, 0, 0)) == (5e-9, 2.5e-9, 0.5e-9)

    def test_centre(self):
        p1 = (-18.5, 10, 0)
        p2 = (10, 5, 10)
        cell = (0.1, 0.25, 2)
        mesh = df.Mesh(p1, p2, cell)
        assert isinstance(mesh.centre(), tuple)
        assert mesh.centre() == (-4.25, 7.5, 5)

        p1 = (500e-9, 125e-9, 3e-9)
        p2 = (0, 0, 0)
        cell = (10e-9, 5e-9, 1e-9)
        mesh = df.Mesh(p1, p2, cell)
        assert isinstance(mesh.centre(), tuple)
        assert mesh.centre() == (250e-9, 62.5e-9, 1.5e-9)

    def test_various_conversions(self):
        p1 = (0, 0, 0)
        p2 = (9, 5, -11)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1, p2, cell)

        assert mesh.point2index(mesh.centre()) == (4, 2, 5)

    def test_random_point(self):
        p1 = (-18.5, 5, 0)
        p2 = (10, -10, 10e-9)
        cell = (0.1e-9, 0.25, 2e-9)
        mesh = df.Mesh(p1, p2, cell)

        for _ in range(20):
            p = mesh.random_point()
            assert isinstance(p, tuple)
            for i in range(3):
                assert mesh.pmin[i] <= p[i] <= mesh.pmax[i]

    def test_plot(self):
        for arg in self.valid_args:
            p1, p2, cell = arg
            mesh = df.Mesh(p1, p2, cell)

            assert isinstance(mesh.plot(), matplotlib.figure.Figure)

    def test_cells(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1, p2, cell)

        counter = 0
        for cell in mesh.cells():
            assert isinstance(cell, tuple)
            assert len(cell) == 2

            i, p = cell
            assert isinstance(i, tuple)
            assert len(i) == 3
            assert isinstance(p, tuple)
            assert len(p) == 3
            for j in range(3):
                assert i[j] >= 0
                assert i[j] <= 9
                assert p[j] >= 0.5
                assert p[j] <= 9.5

            counter += 1

        assert counter == 1000

    def test_line_intersection(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1, p2, cell)

        tol = 1e-12
        li = mesh.line_intersection((1, 1, 1), (5, 5, 5), n=10)
        for point in li:
            assert isinstance(point, tuple)
            assert len(point) == 2

            d, p = point
            assert isinstance(d, float)
            assert 0 <= d <= 10*np.sqrt(3) + tol
            assert isinstance(p, tuple)
            assert len(p) == 3
            for j in range(3):
                assert 0 <= p[j] <= 10

        li = list(mesh.line_intersection((1, 0, 0), (0, 0, 0), n=30))

        assert len(li) == 30
        assert li[0][0] == 0
        assert abs(li[-1][0] - 10) < tol
        assert li[0][1] == (0, 0, 0)
        assert li[-1][1] == (10, 0, 0)

    def test_script(self):
        for arg in self.valid_args:
            p1, p2, cell = arg
            mesh = df.Mesh(p1, p2, cell)

            with pytest.raises(NotImplementedError):
                mesh.script()

    def test_cell_greater_than_mesh_domain(self):
        p1 = (0, 0, 0)
        p2 = (1., 1., 1.)

        for d in [(2, 1, 1), (1, 2, 1), (1, 1, 2), (1, 5, 0.1)]:
            with pytest.raises(ValueError) as excinfo:
                mymesh = df.Mesh(p1, p2, d)

    def test_mesh_domain_not_aggregate_of_cell(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)

        for d in [(0.6, 1, 1), (1, 2.2, 1), (1, 1, 4), (2, 5, 0.8)]:
            with pytest.raises(ValueError) as excinfo:
                mymesh = df.Mesh(p1, p2, d)



import random
import numpy as np
import joommfutil.typesystem as ts
import discretisedfield.util as dfu
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


@ts.typesystem(p1=ts.RealVector(size=3),
               p2=ts.RealVector(size=3),
               cell=ts.PositiveRealVector(size=3),
               name=ts.ObjectName,
               l=ts.PositiveRealVector(size=3),
               pmin=ts.RealVector(size=3),
               pmax=ts.RealVector(size=3),
               n=ts.PositiveIntVector(size=3))
class Mesh(object):
    def __init__(self, p1, p2, cell, name="mesh"):
        """
        Creates a rectangular finite difference mesh.

        Args:
          p1 (tuple, list, np.ndarray): First mesh domain point
            p1 is of length 3 (xmin, ymin, zmax).
          p2 (tuple, list, np.ndarray): Second mesh domain point
            p2 is of length 3 (xmin, ymin, zmax).
          cell (tuple, list, np.ndarray): Discretisation cell size
            cell is of length 3 and defines the discretisation steps in
            x, y, and z directions: (dx, dy, dz).
          name (Optional[str]): Mesh name

        Attributes:
          p1 (tuple): First mesh domain point

          p2 (tuple): Second mesh domain point

          cell (tuple): Discretisation cell size

          name (str): Mesh name

          pmin (tuple): Minimum mesh domain point

          pmax (tuple): Maximum mesh domain point

          l (tuple): length of domain x, y, and z edges (lx, ly, lz):

            lx = abs(p2[0] - p1[0])

            ly = abs(p2[1] - p1[2])

            lz = abs(p2[2] - p1[2])

          n (tuple): The number of cells in three dimensions (nx, ny, nz):

            nx = lx/dx

            ny = ly/dy

            nz = lz/dz

        """
        self.p1 = tuple(p1)
        self.p2 = tuple(p2)
        self.cell = tuple(cell)
        self.name = name

        # Compute domain edge lengths.
        self.l = (abs(self.p2[0]-self.p1[0]),
                  abs(self.p2[1]-self.p1[1]),
                  abs(self.p2[2]-self.p1[2]))

        # Compute minimum and maximum mesh domain points.
        self.pmin = (min(self.p1[0], self.p2[0]),
                     min(self.p1[1], self.p2[1]),
                     min(self.p1[2], self.p2[2]))
        self.pmax = (max(self.p1[0], self.p2[0]),
                     max(self.p1[1], self.p2[1]),
                     max(self.p1[2], self.p2[2]))

        tol = 1e-12  # picometer tolerance
        # Check if the discretisation cell size is greater than the domain.
        for i in range(3):
            if self.cell[i] > self.l[i]:
                raise ValueError(("Discretisation cell is greater than "
                                  "the domain dimension: cell[{}] > "
                                  "abs(p2[{}]-p1[{}]).").format(i, i, i))

        # Check if the domain is not an aggregate of discretisation cell.
        for i in range(3):
            if tol < self.l[i] % self.cell[i] < self.cell[i] - tol:
                raise ValueError(("Domain is not a multiple (aggregate) of "
                                  "the discretisation cell: "
                                  "abs(p2[{}]-p1[{}]) % "
                                  "cell[{}].").format(i, i, i))

        # Compute the number of cells in all three dimensions.
        self.n = (int(round(self.l[0]/self.cell[0])),
                  int(round(self.l[1]/self.cell[1])),
                  int(round(self.l[2]/self.cell[2])))

    def __repr__(self):
        """Mesh representation method.

        Returns:
          A mesh representation string.

        """
        p1str = "p1=({}, {}, {})".format(self.p1[0], self.p1[1], self.p1[2])
        p2str = "p2=({}, {}, {})".format(self.p2[0], self.p2[1], self.p2[2])
        cellstr = "cell=({}, {}, {})".format(self.cell[0],
                                             self.cell[1],
                                             self.cell[2])
        namestr = "name=\"{}\"".format(self.name)

        return "Mesh({}, {}, {}, {})".format(p1str, p2str, cellstr, namestr)

    def _ipython_display_(self):
        """Shows a matplotlib figure of sample range and discretisation."""
        # TODO: plt.show() works only with nbagg
        fig = self.plot()  # pragma: no cover
        plt.show()  # pragma: no cover

    def centre(self):
        """Compute and return the mesh centre point.

        Returns:
          A mesh centre point tuple of coordinates.

        """
        return (self.pmin[0] + 0.5*self.l[0],
                self.pmin[1] + 0.5*self.l[1],
                self.pmin[2] + 0.5*self.l[2])

    def random_point(self):
        """Generate a random mesh point.

        Returns:
          A random mesh point tuple of coordinates.

        """
        return (self.pmin[0] + random.random()*self.l[0],
                self.pmin[1] + random.random()*self.l[1],
                self.pmin[2] + random.random()*self.l[2])

    def index2point(self, i):
        """Convert the discretisation cell index to its centre point coordinate.

        The finite difference domain is disretised in x, y, and z directions
        in dx, dy, and dz steps, respectively. Accordingly, there are
        nx, ny, and nz discretisation steps. This method converts the cell
        index (ix, iy, iz) to the cell's centre point coordinate.

        This method raises ValueError if the index is out of range.

        Args:
          i (tuple): A length 3 tuple of integers (ix, iy, iz)

        Returns:
          A length 3 tuple of x, y, and z coodinates

        """
        for j in range(3):
            if i[j] < 0 or i[j] > self.n[j] - 1:
                raise ValueError(("Index i[{}]={} out of "
                                  "range.").format(j, i[j]))

        return (self.pmin[0] + (i[0]+0.5)*self.cell[0],
                self.pmin[1] + (i[1]+0.5)*self.cell[1],
                self.pmin[2] + (i[2]+0.5)*self.cell[2])

    def point2index(self, p):
        """Compute the index of a cell containing point p.

        This method is an inverse function of index2point method.
        (For details on index, please refer to the index2point method.)

        It raises ValueError if the point is outside the mesh.

        Args:
          p (tuple): A length 3 tuple of Real numbers (px, py, pz)

        Returns:
          A length 3 cell index tuple (ix, iy, iz).

        """
        for j in range(3):
            if p[j] < self.pmin[j] or p[j] > self.pmax[j]:
                raise ValueError(("Point coordinate p[{}]={} outside "
                                 "the mesh domain."). format(j, p[j]))

        i = []
        for j in range(3):
            ij = int(round((p[j]-self.pmin[j])/self.cell[j] - 0.5))

            # If rounded to the out-of-range mesh index.
            if ij < 0:
                ij = 0  # pragma: no cover
            elif ij > self.n[j] - 1:
                ij = self.n[j] - 1

            i.append(ij)

        return tuple(i)

    def cell_centre(self, p):
        """Computes the centre of cell containing (or nearest) to point p.

        Args:
          p (tuple): A length 3 tuple of point coordinates

        Returns:
          A length 3 tuple of cell's centre coordinates

        """
        return self.index2point(self.point2index(p))

    def plot(self):
        """Creates a figure of a mesh range and discretisation cell."""
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.set_aspect("equal")

        cell_point = (self.pmin[0] + self.cell[0],
                      self.pmin[1] + self.cell[1],
                      self.pmin[2] + self.cell[2])

        dfu.plot_box(ax, self.pmin, self.pmax)
        dfu.plot_box(ax, self.pmin, cell_point, props="r-", linewidth=1)

        ax.set(xlabel=r"$x$", ylabel=r"$y$", zlabel=r"$z$")

        return fig

    def cells(self):
        """Generator iterating through all mesh cells and
        yielding mesh indices and centre coordinates."""
        for k in range(self.n[2]):
            for j in range(self.n[1]):
                for i in range(self.n[0]):
                    yield (i, j, k), self.index2point((i, j, k))

    def line_intersection(self, l, l0, n=100):
        """Generator yielding mesh cell indices and their centre coordinates,
        along the line defined with l and l0 in n points."""
        try:
            p1, p2 = dfu.box_line_intersection(self.pmin, self.pmax, l, l0)
        except TypeError:
            raise ValueError("Line does not intersect mesh in two points.")

        p1, p2 = np.array(p1), np.array(p2)
        dl = (p2-p1) / (n-1)
        for i in range(n):
            point = p1 + i*dl
            yield np.linalg.norm(i*dl), tuple(point)

    def script(self):
        """This method should be implemented by a specific
        micromagnetic calculator."""
        raise NotImplementedError



import setuptools

with open("README.rst") as f:
    readme = f.read()

setuptools.setup(
    name="discretisedfield",
    version="0.5.6",
    description=("A Python package for analysing and manipulating "
                 "discretised fields."),
    long_description=readme,
    url='https://joommf.github.io',
    author='Marijan Beg, Ryan A. Pepper, and Hans Fangohr',
    author_email='jupyteroommf@gmail.com',
    packages=setuptools.find_packages(),
    install_requires=["joommfutil",
                      "matplotlib"],
    classifiers=['Development Status :: 3 - Alpha',
                 'License :: OSI Approved :: BSD License',
                 'Topic :: Scientific/Engineering :: Physics',
                 'Intended Audience :: Science/Research',
                 'Programming Language :: Python :: 3 :: Only']
)
